(***************************************************************************)
(*                              MPC 2012                                   *)
(*                                                                         *)
(*                                                                         *)
(* Authors: Louis Mandel (louis.mandel@lri.fr)                             *)
(*          Florence Plateau (florence.plateau@lri.fr)                     *)
(*                                                                         *)
(* Creation date: April 2012                                               *)
(*                                                                         *)
(***************************************************************************)

(** Definition of clock expressions and adaptability constraint systems.

    @author Louis Mandel
    @author Florence Plateau
*)

open Misc
open Pbword

(** {2 Data structures} *)

(** Unknown word variable: [c_n]. *)
type word_variable = C of int

(** Set of word variables. *)
module Word_var_set = Set.Make(struct
  type t = word_variable
  let compare = Pervasives.compare
end)

(** Map of word variables. *)
module Word_var_map = Map.Make(struct
  type t = word_variable
  let compare = Pervasives.compare
end)

(** Choice of number of 1s for a set of word variables *)
type choice_of_number_of_1 =
    { k: int;
      k': int;
      nbones_env: (int * int) Word_var_map.t; }

(** Expression on words: [w ::= c | p | w on w]. *)
type word_expr =
  | Var of word_variable
  | Pbw of periodic_binary_word
  | On of word_expr * word_expr

(** Adaptability constraints:
    [p_1 on ... <: p_2 on ...] and
    [c_x on p_1' <: c_y on p_2'].
*)
type adaptability_cstr = Adaptability of word_expr * word_expr

(** Simplified adaptability constraint: [c_x on p_x <: c_y on p_y]. *)
type simplified_adaptability_cstr =
    Simpl_adaptability of
      (word_variable * periodic_binary_word) *
        (word_variable * periodic_binary_word)

(** Simplified adaptability constraint system. *)
type simplified_adaptability_cstrs = simplified_adaptability_cstr list


(** Synchronizability constraint: [c_x on p_x |><| c_y on p_y]. *)
type synchronizability_cstr =
    Synchronizability of
      (word_variable * periodic_binary_word) *
        (word_variable * periodic_binary_word)

(** Synchronizability constraint system. *)
type synchronizability_cstrs = synchronizability_cstr list


(** Precedence constraint: [c_x on p_x <= c_y on p_y]. *)
type precedence_cstr =
    Precedence of
      (word_variable * periodic_binary_word) *
        (word_variable * periodic_binary_word)

(** Precedence constraint system. *)
type precedence_cstrs = precedence_cstr list


(** Variables corresponding to the size (i.e., [|c_n|]) and the index of
    a 1 (i.e., [I_{c_n}(j)]) of a word variable [c_n].
*)
type integer_variable =
  | Size of word_variable
  | Iof of word_variable * int

(** Ordered integer variables *)
module Ordered_integer_var = struct
  type t = integer_variable
  let compare x1 x2 =
    match x1, x2 with
    | Iof (c1, j1), Iof (c2, j2) ->
        let cmp = Pervasives.compare c1 c2 in
        if cmp = 0 then Pervasives.compare j1 j2
        else cmp
    | Size c1, Size c2 -> Pervasives.compare c1 c2
    | Iof _, Size _ -> -1
    | Size _, Iof _ -> 1
end

(** Set of integer variables. *)
module Integer_var_set = Set.Make(Ordered_integer_var)

(** Map of integer variables. *)
module Integer_var_map = Map.Make(Ordered_integer_var)


(** Integer linear constraints. *)
type linear_cstr =
  | Eq of (int * integer_variable) list * int
  | Le of (int * integer_variable) list * int
  | Ge of (int * integer_variable) list * int

(** Linear constraint system. *)
type linear_cstrs = linear_cstr list


(** Constraints on sizes and indexes of 1s. *)
type iof_and_size_cstrs =
    { choice: choice_of_number_of_1;
      synchronizability: linear_cstrs;
      precedence: linear_cstrs;
      periodicity: linear_cstrs;
      sufficient_size: linear_cstrs;
      sufficient_indexes: linear_cstrs;
      increase_indexes: linear_cstrs; }

(** {2 Functions on clock expressions} *)

(** Compute [on]s in an expression [c on p1 on p2 on ...] or
    [p1 on p2 on ...].
*)
let rec on_computation w =
  match w with
  | Var c -> Some c, one
  | Pbw p -> None, p
  | On (w1, w2) ->
      let c1_opt, p1 = on_computation w1 in
      let c2_opt, p2 = on_computation w2 in
      assert (c2_opt = None);
      c1_opt, on p1 p2

(** {2 Functions on simplified adaptability constraint systems } *)

(** Samplers of a variable c_n in a simplified adaptability constraint
    system A:
    [{ p_n | (c_n on p_n <: c_y on p_y) \in A \/
             (c_x on p_x <: c_n on p_n) \in A }].
*)
let samplers c_n cstrs =
  List.fold_left
    (fun acc (Simpl_adaptability ((c_x, p_x), (c_y, p_y))) ->
      let acc = if c_x = c_n then p_x :: acc else acc in
      if c_y = c_n then p_y :: acc else acc)
    [] cstrs

(** Set of word variables that appear in a simplified adaptability
    constraint system.
*)
let vars_of_simpl_adaptability_cstrs cstrs =
  List.fold_left
    (fun acc (Simpl_adaptability ((c_x, p_x), (c_y, p_y))) ->
      Word_var_set.add c_x (Word_var_set.add c_y acc))
    Word_var_set.empty cstrs


(** {2 Functions on linear constraint systems} *)

(** Set of integer variables that appear in a linear constraint system. *)
let vars_of_linear_constraints cstrs =
  List.fold_left
    (fun acc (Eq (terms, _) | Le (terms, _) | Ge (terms, _)) ->
      List.fold_left
        (fun acc (_, x) -> Integer_var_set.add x acc)
        acc terms)
    Integer_var_set.empty cstrs


(** {2 Printing functions } *)

let fprint_word_var ff (C n) =
  Format.fprintf ff "c_%d" n

let rec fprint_word_expr ff w =
  match w with
  | Var c -> fprint_word_var ff c
  | Pbw p -> fprint_pbw ff p
  | On (w1, w2) ->
      Format.fprintf ff "%a on %a"
        fprint_word_expr w1
        fprint_word_expr w2

let fprint_adaptability_cstr ff (Adaptability (w1, w2)) =
  Format.fprintf ff "@[%a@] <: @[%a@]"
    fprint_word_expr w1
    fprint_word_expr w2

let fprint_adaptability_cstrs ff cstrs =
  Format.fprintf ff "{@[<hov 1> %a@] }"
    (fprint_list ";@\n" fprint_adaptability_cstr) cstrs


let fprint_simpl_adaptability_cstr ff
    (Simpl_adaptability ((c_x, p_x), (c_y, p_y))) =
  Format.fprintf ff "@[%a on %a@] <: @[%a on %a@]"
    fprint_word_var c_x
    fprint_pbw p_x
    fprint_word_var c_y
    fprint_pbw p_y

let fprint_simpl_adaptability_cstrs ff cstrs =
  Format.fprintf ff "{@[<hov 1> %a@] }"
    (fprint_list ";@\n" fprint_simpl_adaptability_cstr) cstrs

let fprint_synchronizability_cstr ff
    (Synchronizability ((c_x, p_x), (c_y, p_y))) =
  Format.fprintf ff "@[%a on %a@] |><| @[%a on %a@]"
    fprint_word_var c_x
    fprint_pbw p_x
    fprint_word_var c_y
    fprint_pbw p_y

let fprint_synchronizability_cstrs ff cstrs =
  Format.fprintf ff "{@[<hov 1> %a@] }"
    (fprint_list ";@\n" fprint_synchronizability_cstr) cstrs

let fprint_precedence_cstr ff (Precedence ((c_x, p_x), (c_y, p_y))) =
  Format.fprintf ff "@[%a on %a@] <= @[%a on %a@]"
    fprint_word_var c_x
    fprint_pbw p_x
    fprint_word_var c_y
    fprint_pbw p_y

let fprint_precedence_cstrs ff cstrs =
  Format.fprintf ff "{@[<hov 1> %a@] }"
    (fprint_list ";@\n" fprint_precedence_cstr) cstrs

let fprint_choice_c_n ff (c_n, (nbones_c_n_u, nbones_c_n_v))=
  Format.fprintf ff "|%a.u|_1 = %i and |%a.v|_1 = %i"
    fprint_word_var c_n
    nbones_c_n_u
    fprint_word_var c_n
    nbones_c_n_v

let fprint_choice ff choice =
  Format.fprintf ff "@[%a@]"
    (fprint_list "@\n" fprint_choice_c_n)
    (Word_var_map.bindings choice.nbones_env)

let fprint_integer_var ff var =
  match var with
  | Size c_n ->
      Format.fprintf ff "|%a.v|" fprint_word_var c_n
  | Iof (c_n, j) ->
      Format.fprintf ff "I_{%a}(%i)" fprint_word_var c_n j

let fprint_linear_cstr ff cstr =
  let fprint_term ff (k, x) =
    Format.fprintf ff "(%i * %a)" k fprint_integer_var x
  in
  match cstr with
  | Eq (terms, n) ->
      Format.fprintf ff "%a = %i"
        (fprint_list " + " fprint_term) terms
        n
  | Le (terms, n) ->
      Format.fprintf ff "%a <= %i"
        (fprint_list " + " fprint_term) terms
        n
  | Ge (terms, n) ->
      Format.fprintf ff "%a >= %i"
        (fprint_list " + " fprint_term) terms
        n

let fprint_linear_cstrs ff cstrs =
  Format.fprintf ff "{@[<hov 1> %a@] }"
    (fprint_list ";@\n" fprint_linear_cstr) cstrs

let fprint_linear_cstrs_sol ff sol =
  Integer_var_map.iter
    (fun x n -> Format.fprintf ff "%a = %d@\n" fprint_integer_var x n)
    sol
