(***************************************************************************)
(*                         Periodic Binary Words                           *)
(*                                                                         *)
(*                                                                         *)
(* Authors: Louis Mandel (louis.mandel@lri.fr)                             *)
(*          Florence Plateau (florence.plateau@lri.fr)                     *)
(*                                                                         *)
(* Creation date: August 2008                                              *)
(*                                                                         *)
(***************************************************************************)

(* $Id: cplex.ml,v 1.1 2012-07-20 11:01:48 mandel Exp $ *)

let glpsol = ref "glpsol"
let glpsol_options = ref [||]
let verbose = ref false

type cmp = Eq | Leq | Geq | Lt

type sign = Plus | Minus

type term = sign * int * string

type objective =
  | Minimize of term list
  | Maximize of term list

type constraint_ =
    term list * cmp * int

type 'a system =
    { env: 'a;
      integer: string list;
      bounds: (int option * string * int option) list;
      objective: objective;
      subject_to: constraint_ list; }

type status =
  | GLP_UNDEF
  | GLP_FEAS
  | GLP_NOFEAS
  | GLP_OPT
  | Empty_constraints

let string_of_sign s =
  match s with
  | Plus -> "+"
  | Minus -> "-"

let print_term fmt (sign, c, x) =
  Format.fprintf fmt "%s %d %s " (string_of_sign sign) c x

let print_constraint =
  let string_of_right_part op c =
    match op with
    | Eq -> "= " ^ (string_of_int c)
    | Leq -> "<= " ^ (string_of_int c)
    | Geq -> ">= " ^ (string_of_int c)
    | Lt -> "<= " ^ (string_of_int (c - 1))
  in
  fun fmt (terms, cmp, c) ->
    List.iter (print_term fmt) terms;
    Format.fprintf fmt "%s" (string_of_right_part cmp c)

let output out sys =
  let objective obj =
    let terms =
      match obj with
      | Minimize terms -> Format.fprintf out "Minimize@\n"; terms
      | Maximize terms -> Format.fprintf out "Maximize@\n"; terms
    in
    match terms with
    | [] -> assert false
    | terms ->
        Format.fprintf out "\t";
        List.iter (print_term out) terms;
        Format.fprintf out "@\n@\n";
  in
  let subject_to cstrs =
    Format.fprintf out "Subject To@\n";
    List.iter
      (fun cstr -> Format.fprintf out "\t%a@\n" print_constraint cstr)
      cstrs;
    Format.fprintf out "@\n@\n"
  in
  let bounds bound_list =
    Format.fprintf out "Bounds@\n";
    match bound_list with
    | [] -> assert false
    | _ ->
        let m_inf = (* "-inf" *) string_of_int min_int in
        let p_inf = (* "+inf" *) string_of_int max_int in
        List.iter
          (function
            | (None, var, None) ->
                Format.fprintf out "\t%s <= %s <= %s@\n" m_inf var p_inf
            | (Some n1, var, None) ->
                Format.fprintf out "\t%d <= %s <= %s @\n" n1 var p_inf
            | (None, var, Some n2) ->
                Format.fprintf out "\t%s <= %s <= %d @\n" m_inf var n2
            | (Some n1, var, Some n2) ->
                Format.fprintf out "\t%d <= %s <= %d @\n" n1 var n2)
          bound_list;
        Format.fprintf out "@\n@\n"
  in
  let integer vars =
    Format.fprintf out "Integer@\n";
    List.iter (fun var -> Format.fprintf out "\t%s@\n" var) vars;
    Format.fprintf out "@\n@\n"
  in
  objective sys.objective;
  subject_to sys.subject_to;
  bounds sys.bounds;
  integer sys.integer;
  Format.fprintf out "end@."


let read_out chan =
  let acc = ref [] in
  let line = ref "" in
  let next_line regexp =
    line := input_line chan;
    try
      let _ = Str.search_forward (Str.regexp regexp) !line 0 in
      ()
    with Not_found ->
      print_endline ("XXX"^(!line)^"XXX");
      assert false
  in
  let get i = Str.matched_group i !line in
  next_line "Problem:";
  next_line "Rows: *\\([0-9]+\\)";
  let rows = Pervasives.int_of_string (get 1) in
  next_line "Columns:[^0-9]*\\([0-9]+\\)";
  let columns = Pervasives.int_of_string (get 1) in
  next_line "Non-zeros:";
  next_line "Status:\\(.*\\)";
(*   let status = get 1 in *)
  next_line "Objective:";
  next_line "";
  next_line "No\\.";
  next_line "---";
  for i = 1 to rows do
    next_line (Pervasives.string_of_int i);
  done;
  next_line "";
  next_line "No\\.";
  next_line "---";
  for i = 1 to columns do
(*assert (next_line ((string_of_int i)^" \\([^ ]+\\) *\\* *\\(-?[0-9]+\\)"));*)
    next_line ((Pervasives.string_of_int i)^" \\([^ ]+\\)\\(.*\\)");
    let var = get 1 in
(*     let sol = int_of_string (get 2) in *)
    let reminder = get 2 in
    if not (Str.string_match (Str.regexp ".*\\*") reminder 0) then
      next_line ".*";
    acc := (i, var) :: !acc;
  done;
  !acc

let read_sol chan =
  let acc = ref [] in
  let line = ref "" in
  let next_line regexp =
    line := input_line chan;
    try
      let _ = Str.search_forward (Str.regexp regexp) !line 0 in
      ()
    with Not_found ->
      print_endline ("#"^(!line)^"#");
      assert false
  in
  let get i = Str.matched_group i !line in
  next_line "\\([0-9]+\\) *\\([0-9]+\\)";
  let rows = Pervasives.int_of_string (get 1) in
  let columns = Pervasives.int_of_string (get 2) in
  next_line "\\([0-9]+\\)";
  let status =
    match Pervasives.int_of_string (get 1) with
    | 1 -> GLP_UNDEF
    | 2 -> GLP_FEAS
    | 4 -> GLP_NOFEAS
    | 5 -> GLP_OPT
    | _ -> assert false
  in
  for i = 1 to rows do
    next_line "";
  done;
  for i = 1 to columns do
    next_line "\\(-?[0-9]+\\)";
    let sol = int_of_string (get 1) in
    acc := (i, sol) :: !acc;
  done;
  status, !acc

let read out_chan sol_chan =
  let col_x_var_list = read_out out_chan in
  let status, col_x_sol_list = read_sol sol_chan in
  status,
  List.fold_left
    (fun acc (i, x) ->
      try (x, List.assoc i col_x_sol_list)::acc
      with Not_found -> assert false)
    [] col_x_var_list


let execute_prog prog args new_stdin new_stdout new_stderr =

  let pid =
    Unix.handle_unix_error
      (Unix.create_process prog args new_stdin new_stdout) new_stderr
  in
  let rec my_waitpid pid =
    try Unix.waitpid [] pid
    with Unix.Unix_error (Unix.EINTR, _, _) -> my_waitpid pid
    | Unix.Unix_error (err, _, _) ->
        Format.eprintf "waitpid: %s@." (Unix.error_message err);
        exit 2
  in
  let _, status = my_waitpid pid in
  match status with
  | Unix.WEXITED n -> Some n
  | Unix.WSIGNALED n ->
      Format.eprintf "Warning: %s was killed by a signal %i.@." prog n;
      None
  | Unix.WSTOPPED n ->
      Format.eprintf "Warning: %s was stopped by a signal %i.@." prog n;
      None

let int_of_sign_int sign k =
  match sign with
  | Plus -> k
  | Minus -> - k

let sign_int_of_int n =
  if n >= 0 then
    Plus, n
  else
    Minus, - n

let add sign1 k1 sign2 k2 =
  sign_int_of_int (int_of_sign_int sign1 k1 + int_of_sign_int sign2 k2)

(* combines multiple occurrence of the same variable in a term *)
let simplify_term (terms : term list) =
  List.rev
    (List.fold_left
       (fun res (sign, k, x) ->
         let b, res' =
           List.fold_left
             (fun (b, acc) ((sign', k', x') as t) ->
               if Pervasives.(=) x x' then
                 let sign'', k'' = add sign k sign' k' in
                 (true, (sign'', k'', x) :: acc)
               else
                 (b, t :: acc))
             (false, []) res
         in
         let res' = List.rev res' in
         if b then res' else (sign, k, x) :: res')
       [] terms)

(* combines multiple occurrence of the same variable in a constraint *)
let simplify_constraint cstr =
  let terms, cmp, n = cstr in
  let terms' = simplify_term terms in
  terms', cmp, n

let simplify_subject_to cstrs =
  List.rev (List.rev_map simplify_constraint cstrs)

let simplify_objective obj =
  match obj with
  | Minimize terms -> Minimize (simplify_term terms)
  | Maximize terms -> Maximize (simplify_term terms)

let simplify sys =
  { sys with subject_to = simplify_subject_to sys.subject_to;
             objective = simplify_objective sys.objective; }

let solve sys =
  let sys = simplify sys in
  match sys.subject_to with
  | [] ->
      Empty_constraints,
      List.map
        (function
          | (None, var, None) -> assert false
          | (Some n1, var, None) -> var, n1
          | (None, var, Some n2) -> var, n2
          | (Some n1, var, Some n2) ->
              begin match sys.objective with
              | Minimize _ -> var, n1
              | Maximize _ -> var, n2
              end)
        sys.bounds
  | _ ->
      let sys_file, out_ch = Filename.open_temp_file "sys_" ".lp" in
      let out = Format.formatter_of_out_channel out_ch in
      let log_file = (Filename.chop_suffix sys_file ".lp")^".log" in
      let log_fd =
        Unix.openfile log_file
          [ Unix.O_WRONLY; Unix.O_CREAT; Unix.O_TRUNC ] 0o640
      in
      let out_file = (Filename.chop_suffix sys_file ".lp")^".out" in
      let sol_file = (Filename.chop_suffix sys_file ".lp")^".sol" in
      if !verbose then begin
        Format.printf "Problem: %s@\n" sys_file;
        Format.printf "Log: %s@\n" log_file;
        Format.printf "Output: %s@\n" out_file;
        Format.printf "Solution: %s@\n" sol_file;
      end;
      output out sys;
      close_out out_ch;
      begin match
(*
        Sys.command (!glpsol^" --cpxlp -o "^out_file^
                     " -w "^sol_file^
                     " --log"^log_file^" "^sys_file)
*)
        execute_prog !glpsol
          (Array.concat
             [ [| !glpsol |];
               !glpsol_options;
               [| "--cpxlp";
                  "-o"; out_file;
                  "-w"; sol_file;
                  "--log"; log_file;
                  sys_file |] ])
          Unix.stdin log_fd Unix.stderr
      with
      | Some 0 ->
          Unix.close log_fd;
          let out_chan = open_in out_file in
          let sol_chan = open_in sol_file in
          let status, res = read out_chan sol_chan in
          close_in out_chan;
          close_in sol_chan;
          if not !verbose then begin
            (try Sys.remove sys_file with Sys_error _ -> ());
            (try Sys.remove out_file with Sys_error _ -> ());
            (try Sys.remove sol_file with Sys_error _ -> ())
          end;
          status, res
      | None ->
          Unix.close log_fd;
          if not !verbose then begin
            (try Sys.remove sys_file with Sys_error _ -> ());
            (try Sys.remove out_file with Sys_error _ -> ());
            (try Sys.remove sol_file with Sys_error _ -> ())
          end;
          GLP_NOFEAS, []
      | Some _ ->
          Format.eprintf "Fail to execute: ";
          let cmd =
            Array.concat
              [ [| !glpsol |];
                !glpsol_options;
                [| "--cpxlp";
                   "-o"; out_file;
                   "-w"; sol_file;
                   "--log"; log_file;
                   sys_file |] ]
          in
          for i = 0 to Pervasives.(-) (Array.length cmd) 1 do
            Format.eprintf "%s " cmd.(i);
          done ;
          Format.eprintf "@.";
          Format.eprintf "%s must be installed.@." !glpsol;
          exit 1
      end

