(***************************************************************************)
(*                              MPC 2012                                   *)
(*                                                                         *)
(*                                                                         *)
(* Authors: Louis Mandel (louis.mandel@lri.fr)                             *)
(*          Florence Plateau (florence.plateau@lri.fr)                     *)
(*                                                                         *)
(* Creation date: April 2012                                               *)
(*                                                                         *)
(***************************************************************************)

(** Naive implementation of ultimately Periodic Binary Words.

    @author Louis Mandel
    @author Florence Plateau
*)

open Misc

(** {2 Data structures} *)

(** Finite word:
    [u ::= epsilon | 0u | 1v].
*)
type word = bool list

(** Ultimately Periodic Binary Words:
    [p ::= u(v)].
*)
type periodic_binary_word = { u: word; v: word; }


(** The ultimately periodic binary word [(1)]. *)
let one = { u = []; v =[ true ]; }


(** {2 Functions on ultimately periodic binary words} *)

(** Size of a finite word [u]:
    [|u|].
*)
let size u = List.length u

(** Number of 1s a finite word [u]:
    [|u|_1].
*)
let nbones u =
  List.fold_left (fun cpt b -> if b then cpt + 1 else cpt) 0 u

(** Index of the jth 1 of an ultimately periodic binary word [p]:
    [I_p(j)].
*)
let iof p j =
  assert (j >= 0);
  let rec iof w j =
    match w with
    | [] -> iof p.v j
    | false :: w' -> 1 + iof w' j
    | true :: w' -> if j = 1 then 1 else 1 + iof w' (j - 1)
  in
  if j = 0 then 0
  else iof p.u j

(** Increase prefix size of [n] bits:
    [u(vv') = uv(v'v)] with [|v| = n].
*)
let rec increase_prefix p n =
  assert (n >= 0);
  if n = 0 then p
  else
    match p.v with
    | [] -> assert false
    | b :: v -> increase_prefix { u = p.u @ [b]; v = p.v @ [b]; } (n - 1)

(** Repeat periodic pattern: [u(v) = u(v^k)]. *)
let repeat_periodic_pattern p k =
  assert (k > 0);
  let rec repeat v k =
    if k = 1 then v
    else repeat (p.v @ v) (k - 1)
  in
  { u = p.u; v = repeat p.v k; }

(** Computation of the rate of an ultimately periodic binary word:
    [rate p = |p.v|_1 / |p.v|].
*)
let rate p = (nbones p.v, size p.v)

(** The synchronizability test of two ultimately periodic binary words:
    [p_1 |><| p_2 <=> rate p1 = rate p2].
*)
let ( |><| ) p_1 p_2 =
  let r1 = rate p_1 in
  let r2 = rate p_2 in
  let qred (num, den) =
    let g = gcd num den in
    let num = num / g in
    let den = den / g in
    let sign = den / abs den in
    (sign * num, abs den)
  in
  qred r1 = qred r2

(** The precedence test of two ultimately periodic binary words:
    [p_1 <= p_2 <=> I_{p_1}(j) <= I_{p_2}(j)],
    for all [1 <= j <= h]
    with [h = max (|p_1.u|_1,|p_2.u|_1) + lcm (|p_1.v|_1,|p_2.v|_1)].
*)
let prece p_1 p_2 =
  let h =
    max (nbones p_1.u) (nbones p_2.u) + lcm (nbones p_1.v) (nbones p_2.v)
  in
  let rec check j =
    if j <= h then
      (iof p_1 j <= iof p_2 j) && check (j + 1)
    else
      true
  in
  check 1


(** The adaptability test of two ultimately periodic binary words:
    [p_1 <: p_2].
*)
let ( <: ) p1 p2 = p1 |><| p2 && prece p1 p2


(** Computation of the [on] operator:
    [u_3(v_3) = u_1(v_1) on u_2(v_2)]
    with [|u_3| = max (|u_1|, I_{p_1}(|u_2|))],
         [|v_3| = (lcm (|v_1|_1, |v_2|) / |v_1|_1) * |v_1|]
    and for all [1 <= i <= |u_3|], [u_3[i] = (p_1 on p_2)[i]]
    and for all [1 <= i <= |v_3|], [v_3[i] = (p_1 on p_2)[|u_3| + i]]
*)
let on p_1 p_2 =
  let u1, v1 = p_1.u, p_1.v in
  let u2, v2 = p_2.u, p_2.v in
  let size_u3 = max (size u1) (iof p_1 (size u2)) in
  let size_v3 = (lcm (nbones v1) (size v2) / (nbones v1)) * size v1 in
  let rec on w1 w2 i =
    if i = 0 then [], (w1, w2)
    else
      match w1, w2 with
      | [], _ -> on v1 w2 i
      | _, [] -> on w1 v2 i
      | false :: w1, w2 ->
          let w3, reminder = on w1 w2 (i-1) in
          false :: w3, reminder
      | true :: w1, b :: w2 ->
          let w3, reminder = on w1 w2 (i-1) in
          b :: w3, reminder
  in
  let u3, (w1, w2) = on u1 u2 size_u3 in
  let v3, (_, _)  = on w1 w2 size_v3 in
  { u = u3; v = v3; }


(** {2 Printing functions} *)

let fprint_bool ff b =
  Format.fprintf ff "%d" (if b then 1 else 0)

let fprint_word ff u =
  fprint_list "" fprint_bool ff u

let fprint_pbw ff p =
  Format.fprintf ff "%a(%a)"
    fprint_word p.u
    fprint_word p.v

