(***************************************************************************)
(*                              MPC 2012                                   *)
(*                                                                         *)
(*                                                                         *)
(* Authors: Louis Mandel (louis.mandel@lri.fr)                             *)
(*          Florence Plateau (florence.plateau@lri.fr)                     *)
(*                                                                         *)
(* Creation date: April 2012                                               *)
(*                                                                         *)
(***************************************************************************)

(** Naive implementation of the resolution algorithm of adaptability
    constraints. The goal of the implementation is to illustrate the
    algorithm presented in the article MPC 2012, not to provide an
    efficient implementation of the algorithm.

    @author Louis Mandel
    @author Florence Plateau
*)

open Misc
open Pbword
open Clock_type
open Clock_expression

(* ======================================================================= *)
(** {2 Resolution algorithm} *)

(** The [Solve] exception is raised if a constraint system cannot be
    solved.
*)
exception Solve

(** {4 Translation of subtyping constraints to adaptability constraints}

    Introduction of word variables [c_n] and simplification of type
    varaibles such that the subtyping constraint system is rewritten
    into an equivalent adaptability constraint system.
*)
(*---------------------------------------*)
(* S1 -> S2                              *)
(*---------------------------------------*)

(** Separate the type variable and the word expression in a clock type. *)
let rec split_clock_type ct =
  match ct with
  | CT_var alpha -> alpha, Pbw one
  | CT_on (CT_var alpha, p) -> alpha, Pbw p
  | CT_on (ct', p) ->
      let alpha, w = split_clock_type ct' in
      alpha, On (w, Pbw p)

(** Rewrite a subtyping constraint
    ['a_x on p_x on p_x' on ... <: 'a_y on p_y on p_y' on ...] into
    an adaptability constraint
    [c_x on p_x on p_x' on ... <: c_y on p_y on p_y' on ...] or into
    [p_x on p_x' on ... <: p_y on p_y' on ...] if ['a_x = 'a_y].
*)
let adaptability_cstr_of_subtyping_cstr (Subtype (ct_x, ct_y)) =
  let Alpha x, w_x = split_clock_type ct_x in
  let Alpha y, w_y = split_clock_type ct_y in
  if x = y then Adaptability (w_x, w_y)
  else Adaptability (On (Var (C x), w_x), On (Var (C y), w_y))

(** Rewrite a subtyping constraint system into an adaptability
    constraint system.
*)
let adaptability_cstrs_of_subtyping_cstrs cstrs =
  List.map adaptability_cstr_of_subtyping_cstr cstrs

(** {4 Simplification of adaptability constraints}

    Computation of [on] and simplification of [p_x <: p_y] constraints
    in an adaptability constraint system. The resulting system is an
    equivalent adaptability constraint system where all the
    constraints have the simple shape [c_x on p_x <: c_y on p_y].
*)
(*---------------------------------------*)
(* S2 -> S3                              *)
(*---------------------------------------*)

(** Simplify an adaptability constraint
    [c_x on p_1 on p_1' on ... <: c_y on p_2 on p_2' on ...] into
    [c_x on p_x <: c_y on p_y] or check that an adaptability constraint
    [p_1 on p_1' on ... <: p_2 on p_2' on ...] is satisfied.
*)
let simplify_adaptability_cstr (Adaptability (w_x, w_y)) =
  match on_computation w_x, on_computation w_y with
  | (None, p_x), (None, p_y) ->
      if p_x <: p_y then None
      else raise Solve
  | (Some c_x, p_x), (Some c_y, p_y) when c_x = c_y ->
      if p_x <: p_y then None
      else raise Solve
  | (Some c_x, p_x), (Some c_y, p_y) ->
      Some (Simpl_adaptability ((c_x, p_x), (c_y, p_y)))
  | (Some _, _), (None, _) | (None, _), (Some _, _) -> assert false


(** Rewrite an adaptability constraint system into an equivalent one
    where all the constraints have the simple shape
    [c_x on p_x <: c_y on p_y].
*)
let simplified_adaptability_cstrs_of_adaptability_cstrs cstrs =
  List.fold_right
    (fun cstr acc ->
      match simplify_adaptability_cstr cstr with
      | None -> acc
      | Some cstr' -> cstr' :: acc)
    cstrs []


(** {4 Translation of simplified adaptability constraints to
       adjusted adaptability constraints}

    This step takes as input a simplified adaptability constraint
    system where the shape of the constraint is [c_x on p_x <: c_y on p_y].
    It returns an equivalent adaptability constraint system where for
    each c_n, the size of its samplers is equal.
*)
(*---------------------------------------*)
(* S3 -> S4                              *)
(*---------------------------------------*)


(** Rewrite [c on p] into [c on p'] such that [p = p'] and
    [|p'.u| = max { |p''.u| | p'' \in samplers_c(A) }] and
    [|p'.v| = lcm { |p''.v| | p'' \in samplers_c(A) }].
*)
let adjust_size (c, p) cstrs =
  let samplers_c = samplers c cstrs in
  let size_p'_u, size_p'_v =
    List.fold_left
      (fun (size_u, size_v) p'' ->
        (max size_u (size p''.u), lcm size_v (size p''.v)))
      (size p.u, size p.v) samplers_c
  in
  let p' =
    repeat_periodic_pattern
      (increase_prefix p (size_p'_u - size p.u))
      (size_p'_v / size p.v)
  in
  (c, p')

(** Adjuste the size of each sampler in each constraint of a
    simplified adaptability constraint system.
*)
let adjusted_adaptability_cstrs_of_simplified_adaptability_cstrs cstrs =
  List.map
    (fun (Simpl_adaptability ((c, p1), (c2, p2))) ->
      Simpl_adaptability (adjust_size (c, p1) cstrs,
                          adjust_size (c2, p2) cstrs))
    cstrs


(** {4 Choice of the number of 1s in the solution and
       translation of adjusted adaptability constraints to
       synchronizability and precedence constraints}

    In this step, we choose the number of 1s in each variable [c_n].
    Then we transforme an adjusted adaptability constraint
    system into the corresponding synchronizability and
    precedence constraints.
*)
(*---------------------------------------*)
(* S4 -> S5                              *)
(*---------------------------------------*)

(** Choose for a variable c_n the number of 1s in the prefix
    ([|c_n.u|_1]) and in the periodic part ([|c_n.v|_1]) such that
    [|c_n.u|_1 = |p_n.u| + k * |p_n.v| ( = |p'_n.u| + k * |p'_n.v| = ...)] and
    [|c_n.n|_1 = k' * |p_n.v| ( = k' * |p'_n.v| = ...)]
    with [p_n \in samplers_{c_n}(A)], [p'_n \in samplers_{c_n}(A)], ...
*)
let choose_nbones c_n cstrs k k' =
  match samplers c_n cstrs with
  | [] -> assert false
  | p_n :: samplers_c_n ->
      assert (List.for_all (fun p'_n -> size p_n.u = size p'_n.u) samplers_c_n);
      assert (List.for_all (fun p'_n -> size p_n.v = size p'_n.v) samplers_c_n);
      let nbones_c_n_u = size p_n.u + k * size p_n.v in
      let nbones_c_n_v = k' * size p_n.v in
      (nbones_c_n_u, nbones_c_n_v)

(** Choose for each variable in an adjusted adaptability constraint
    system its number of 1s in the prefix and periodic part.
*)
let make_choice k k' cstrs =
  assert (k >= 0);
  assert (k' > 0);
  let vars = vars_of_simpl_adaptability_cstrs cstrs in
  let nbones_env =
    Word_var_set.fold
      (fun c_n acc ->
        Word_var_map.add c_n (choose_nbones c_n cstrs k k') acc)
      vars Word_var_map.empty
  in
  { k = k;
    k' = k';
    nbones_env = nbones_env; }

(** Split an adaptability constraint [c_x on p_x <: c_y on p_y] into
    a synchronizability constraint [c_x on p_x |><| c_y on p_y] and
    a precedence constraint [c_x on p_x <= c_y on p_y].
*)
let split_adaptability_cstr (Simpl_adaptability ((c_x, p_x), (c_y, p_y))) =
  (Synchronizability ((c_x, p_x), (c_y, p_y)),
   Precedence ((c_x, p_x), (c_y, p_y)))

(** Split an adaptability constraint system into equivalent
    synchronizability and precedence constraint systems.
*)
let synchronizability_and_precedence_cstrs_of_adjusted_adaptability_cstrs
    adaptability_cstrs =
  List.fold_right
    (fun adaptability_cstr (synchronizability_cstrs, precedence_cstrs) ->
      let synchronizability_cstr, precedence_cstr =
        split_adaptability_cstr adaptability_cstr
      in
      (synchronizability_cstr :: synchronizability_cstrs,
       precedence_cstr :: precedence_cstrs))
    adaptability_cstrs ([], [])


(** {4 Synchronizability and precedence constraints to
       linear constraints on the index of 1s and size of unknown words }

    This translation depends on the choice of number on 1s in the
    unknown words.
*)
(*---------------------------------------*)
(* S5 -> S6                              *)
(*---------------------------------------*)

(** Simplify a synchronizability constraint [c_x on p_x |><| c_y on p_y] into
    a constraint on the size of the word variables:
    [|p_y.v|_1 * |c_x| = |p_x.v|_1 * |c_y|].
*)
let simplify_synchronizability_cstr
    (Synchronizability ((c_x, p_x), (c_y, p_y))) =
  Eq ([ (nbones p_y.v, Size c_x); (- nbones p_x.v, Size c_y); ], 0)

(** Simplify a synchronizability constraint system. *)
let simplify_synchronizability_cstrs cstrs =
  List.map simplify_synchronizability_cstr cstrs

(** Simplify a precedence constraint [c_x on p_x <= c_y on p_y] into
    constraints on the indexes of 1s:
    for all [1 <= j <= h],
    [I_{c_x}(I_{p_x}(j)) <= I_{c_x}(I_{p_x}(j))]
    with [h = max (|p_x.u|_1 + k * |p_x.v|_1, |p_y.u|_1 + k * |p_y.v|_1) +
              lcm (k' * |p_x.v|_1, k' * |p_y.v|_1)].
*)
let simplify_precedence_cstr k k'
    (Precedence  ((c_x, p_x), (c_y, p_y))) =
  let h =
    max (nbones p_x.u + k * nbones p_x.v) (nbones p_y.u + k * nbones p_y.v)
      + lcm (k' * nbones p_x.v) (k' * nbones p_y.v)
  in
  let rec build j cstrs =
    if j <= h then
      let cstr =
        Le ([  1, Iof(c_x, iof p_x j);
              -1, Iof(c_y, iof p_y j); ],
            0)
      in
      build (j + 1) (cstr :: cstrs)
    else cstrs
  in
  List.rev (build 1 [])

(** Simplify a precedence constraint system. *)
let simplify_precedence_cstrs choice cstrs =
  List.fold_right
    (fun cstr acc ->
      simplify_precedence_cstr choice.k choice.k' cstr @ acc)
    cstrs []


(** Build the periodicity constraint associated to a variable [I_{c_n}(j')].
    Suppose that [j' = j + l * |c_n.v|_1], the periodicity constraint is:
    [I_{c_n}(j') - I_{c_n}(j) = l * |c_n.v|].
*)
let rec build_periodicity_cstr iof_c_n_j' (nbones_c_n_u, nbones_c_n_v) =
  match iof_c_n_j' with
  | Size _ -> assert false
  | Iof (c_n, j') ->
      assert (j' > nbones_c_n_u + nbones_c_n_v);
      let j'_v = j' - nbones_c_n_u in
      let j_v =
        if j'_v mod nbones_c_n_v = 0 then nbones_c_n_v
        else j'_v mod nbones_c_n_v
      in
      let j = nbones_c_n_u +  j_v in
      let l =
        if j'_v mod nbones_c_n_v = 0 then (j'_v - nbones_c_n_v) / nbones_c_n_v
        else j'_v / nbones_c_n_v
      in
      assert (j' = j + l * nbones_c_n_v);
      Eq ([  1, Iof (c_n, j');
            -1, Iof (c_n, j);
            -l, Size c_n; ], 0)

(** Build the set of periodicity constraints. *)
let build_periodicity_cstrs prec_cstrs choice =
  let iof_vars = vars_of_linear_constraints prec_cstrs in
  let periodicity_cstrs =
    Integer_var_set.fold
      (fun var per_cstrs ->
        match var with
        | Iof (c_n, iof_p_n_j) ->
            let nbones_c_n_u, nbones_c_n_v =
              Word_var_map.find c_n choice.nbones_env
            in
            if iof_p_n_j > nbones_c_n_u + nbones_c_n_v then
              build_periodicity_cstr var (nbones_c_n_u, nbones_c_n_v) ::
              per_cstrs
            else
              per_cstrs
        | _ -> assert false)
      iof_vars []
  in
  List.rev periodicity_cstrs


(** Build sufficient size constraint from the choice of number of 1s of a
    word variable:
    [1 + I_{c_n}(|c_n.u|_1 + |c_n.u|_1) - I_{c_n}(|c_n.u|_1 + 1) <= |c_n.v|].
*)
let build_sufficient_size_cstr c_n (nbones_c_n_u, nbones_c_n_v) =
  Ge
    ([  1, Size c_n ;
       -1, Iof (c_n, nbones_c_n_u + nbones_c_n_v) ;
        1, Iof (c_n, nbones_c_n_u + 1) ; ],
     1)

(** Build the set of sufficient size constraints. *)
let build_sufficient_size_cstrs choice =
  let size_cstrs =
    Word_var_map.fold
      (fun c_n nbones_c_n acc ->
        build_sufficient_size_cstr c_n nbones_c_n :: acc)
      choice.nbones_env []
  in
  List.rev size_cstrs


(** Build sufficient indexes constraints from and index on 1 variable
    [I_{c_n}(j)]:
    [I_{c_n}(j) >= j].
*)
let build_sufficient_index_cstr iof_c_n_j =
  match iof_c_n_j with
  | Size _ -> assert false
  | Iof (c_n, j) ->
      Ge ([ 1, Iof (c_n, j); ], j)

(** Build the set of sufficient indexes constraints. *)
let build_sufficient_index_cstrs prec_cstrs per_cstrs size_cstrs =
  let ( ++ ) = Integer_var_set.union in
  let vars_prec = vars_of_linear_constraints prec_cstrs in
  let vars_per = vars_of_linear_constraints per_cstrs in
  let vars_size = vars_of_linear_constraints size_cstrs in
  let init =
    Integer_var_set.fold
      (fun var acc ->
        match var with
        | Size _ -> acc
        | Iof _ -> build_sufficient_index_cstr var :: acc)
      (vars_prec ++ vars_per ++ vars_size) []
  in
  List.rev init

(** Build increasing indexes constraint:
    [I_{c_n}(j') - I_{c_n}(j) >= j' - j].
*)
let build_increasing_index_cstrs_of_c_n iof_c_n_vars =
  let _, incr_cstrs =
    match Integer_var_set.min_elt iof_c_n_vars with
    | Size _ -> assert false
    | Iof (c_n, j) ->
        Integer_var_set.fold
          (fun var ((c_n, j), acc) ->
            match var with
            | Iof (c_n', j') when c_n = c_n' && j <> j' ->
                assert (j < j');
                ((c_n', j'),
                 Ge ([  1, Iof (c_n, j');
                       -1, Iof (c_n, j); ],
                     j' - j) :: acc)
            | Iof (c_n', j') when c_n = c_n' && j = j' -> ((c_n', j'), acc)
            | _ -> assert false)
          iof_c_n_vars ((c_n, j), [])
  in
  List.rev incr_cstrs

(** Build increasing indexes constraint system. *)
let build_increasing_index_cstrs prec_cstrs per_cstrs size_cstrs =
  let ( ++ ) = Integer_var_set.union in
  let vars_prec = vars_of_linear_constraints prec_cstrs in
  let vars_per = vars_of_linear_constraints per_cstrs in
  let vars_size = vars_of_linear_constraints size_cstrs in
  let iof_vars =
    Integer_var_set.filter
      (function
        | Iof _ -> true
        | Size _ -> false)
      (vars_prec ++ vars_per ++ vars_size)
  in
  let rec iter iof_vars =
    if iof_vars = Integer_var_set.empty then []
    else
      match Integer_var_set.min_elt iof_vars with
      | Size _ -> assert false
      | Iof (c_n, _) ->
          let iof_c_n_vars, iof_others_vars =
            Integer_var_set.partition
              (fun var ->
                match var with
                | Iof (c_n', _) when c_n' = c_n -> true
                | _ -> false)
              iof_vars
          in
          build_increasing_index_cstrs_of_c_n iof_c_n_vars @
          iter iof_others_vars
  in
  iter iof_vars

(** Build indexes of 1s and size constraints corresponding to
    synchronizability and precedence constraints.
*)
let iof_and_size_cstrs_of_synchronizability_and_precedence_cstrs
    choice
    synchronizability_cstrs
    precedence_cstrs =
  let synchronizability =
    simplify_synchronizability_cstrs synchronizability_cstrs
  in
  let precedence =
    simplify_precedence_cstrs choice precedence_cstrs
  in
  let periodicity =
    build_periodicity_cstrs precedence choice
  in
  let sufficient_size =
    build_sufficient_size_cstrs choice
  in
  let sufficient_indexes =
    build_sufficient_index_cstrs precedence periodicity sufficient_size
  in
  let increase_indexes =
    build_increasing_index_cstrs precedence periodicity sufficient_size
  in
  { choice = choice;
    synchronizability = synchronizability;
    precedence = precedence;
    periodicity = periodicity;
    sufficient_size = sufficient_size;
    sufficient_indexes = sufficient_indexes;
    increase_indexes = increase_indexes; }

(* ======================================================================= *)
(** {3 From adaptability constraints to linear constraints} *)

(** Build directely the indexes of 1s and size constraints corresponding to
    adaptability constraints.
*)
let iof_and_size_cstrs_of_adaptability_cstrs ?(verbose = false) k k'
    adaptability_cstrs =
  let simpl_adaptability_cstrs =
    simplified_adaptability_cstrs_of_adaptability_cstrs adaptability_cstrs
  in
  if verbose then begin
    Format.printf "Simplified adaptability constraints:@\n  @[%a@]@\n"
      fprint_simpl_adaptability_cstrs simpl_adaptability_cstrs
  end;
  let adjusted_cstrs =
    adjusted_adaptability_cstrs_of_simplified_adaptability_cstrs
      simpl_adaptability_cstrs
  in
  if verbose then begin
    Format.printf "Adjusted adaptability constraints:@\n  @[%a@]@\n"
      fprint_simpl_adaptability_cstrs adjusted_cstrs
  end;
  let choice = make_choice k k' adjusted_cstrs in
  if verbose then begin
    Format.printf "Choice of number of ones (k = %d and k' = %d):@\n  @[%a@]@\n"
      choice.k
      choice.k'
      fprint_choice choice
  end;
  let synchronizability_cstrs, precedence_cstrs =
    synchronizability_and_precedence_cstrs_of_adjusted_adaptability_cstrs
      adjusted_cstrs
  in
  if verbose then begin
    Format.printf "Synchronizability and precedence constraints:@\n";
    Format.printf "  @[%a@\nand@\n%a@]@\n"
      fprint_synchronizability_cstrs synchronizability_cstrs
      fprint_precedence_cstrs precedence_cstrs
  end;
  let iof_and_size_cstrs =
    iof_and_size_cstrs_of_synchronizability_and_precedence_cstrs
      choice synchronizability_cstrs precedence_cstrs
  in
  iof_and_size_cstrs

(* ======================================================================= *)
(** {2 Linear constraints solving} *)

(** Solve the linear constraints on the indexes of 1s and sizes. *)
let iof_and_size_cstrs_solving ?(verbose = false) cstrs =
  let add_binding x x' vars =
    if Integer_var_map.mem x vars then vars
    else Integer_var_map.add x x' vars
  in
  let cplex_var_of_integer_var x =
    match x with
    | Size (C n) -> Format.sprintf "size_c%d" n
    | Iof (C n, j) -> Format.sprintf "I_c%d_%d" n j
  in
  let sign_of_int n = if n < 0 then Cplex.Minus else Cplex.Plus in
  let cplex_term_of_linear_term (k, x) =
    let x' = cplex_var_of_integer_var x in
    (x, x'), (sign_of_int k, abs k, x')
  in
  let cplex_cmp_of_linear_cstr cstr =
    match cstr with
    | Eq _ -> Cplex.Eq
    | Le _ -> Cplex.Leq
    | Ge _ -> Cplex.Geq
  in
  let cplex_cstr_of_linear_cstr vars cstr =
    match cstr with
    | Eq (terms, c)
    | Le (terms, c)
    | Ge (terms, c) ->
        let vars, terms' =
          List.fold_left
            (fun (vars, terms') term ->
              let (x, x'), term' = cplex_term_of_linear_term term in
              add_binding x x' vars, term' :: terms')
            (vars, [])
            terms
        in
        vars, (List.rev terms', cplex_cmp_of_linear_cstr cstr, c)
  in
  let cplex_cstr_of_linear_cstrs vars cstrs =
    let vars, cplex_cstrs =
      List.fold_left
        (fun (vars, cplex_cstrs) cstr ->
          let vars, cplex_cstr = cplex_cstr_of_linear_cstr vars cstr in
          vars, cplex_cstr :: cplex_cstrs)
        (vars, []) cstrs
    in
    vars, List.rev cplex_cstrs
  in
  let vars = Integer_var_map.empty in
  let vars, synchronizability =
    cplex_cstr_of_linear_cstrs vars cstrs.synchronizability
  in
  let vars, precedence =
    cplex_cstr_of_linear_cstrs vars cstrs.precedence
  in
  let vars, periodicity =
    cplex_cstr_of_linear_cstrs vars cstrs.periodicity
  in
  let vars, sufficient_size =
    cplex_cstr_of_linear_cstrs vars cstrs.sufficient_size
  in
  let vars, sufficient_indexes =
    cplex_cstr_of_linear_cstrs vars cstrs.sufficient_indexes
  in
  let vars, increase_indexes =
    cplex_cstr_of_linear_cstrs vars cstrs.increase_indexes
  in
  let integer =
    Integer_var_map.fold (fun _ x acc -> x :: acc) vars []
  in
  let bounds =
    Integer_var_map.fold (fun _ x acc -> (Some 0, x, None) :: acc) vars []
  in
  let objective =
    let l =
      Integer_var_map.fold
        (fun x x' acc ->
          match x with
          | Iof _ -> (Cplex.Plus, 1, x') :: acc
          | Size _ -> acc) vars []
    in
    Cplex.Minimize l
  in
  let subject_to =
    synchronizability @
    precedence @
    periodicity @
    sufficient_size @
    sufficient_indexes @
    increase_indexes
  in
  let cplex_system =
    { Cplex.env = vars;
      integer = integer;
      bounds = bounds;
      objective = objective;
      subject_to = subject_to;
    }
  in
  let cplex_verbose = !Cplex.verbose in
  Cplex.verbose := verbose;
  if verbose then begin
    Format.printf "Indexes of 1s and size constraints solving:@\n  @[";
  end;
  match Cplex.solve cplex_system with
  | Cplex.GLP_OPT, cplex_sol ->
      Cplex.verbose := cplex_verbose;
      let sol =
        Integer_var_map.map
          (fun x' -> List.assoc x' cplex_sol)
          vars
      in
      if verbose then begin
        Format.printf "%a"
          fprint_linear_cstrs_sol sol;
        Format.printf "@]@\n"
      end;
      sol
  | Cplex.GLP_NOFEAS, _ -> raise Solve
  | Cplex.Empty_constraints, _ -> Integer_var_map.empty
  | _ -> assert false


(** Check the solution of a system of constraints on indexes of 1s and sizes.
*)
let check_iof_and_size_sol iof_and_size_cstrs iof_and_size_sol =
  let check_linear_cstr cstr sol =
    let eval terms =
      List.fold_left
        (fun acc (k, x) ->
          let v =
            try Integer_var_map.find x sol
            with Not_found -> assert false
          in
          acc + k * v)
        0
        terms
    in
    match cstr with
    | Eq (terms, n) -> eval terms = n
    | Le (terms, n) -> eval terms <= n
    | Ge (terms, n) -> eval terms >= n
  in
  let check_linear_cstrs cstrs sol =
    List.for_all (fun cstr -> check_linear_cstr cstr sol) cstrs
  in
  check_linear_cstrs iof_and_size_cstrs.synchronizability iof_and_size_sol &&
  check_linear_cstrs iof_and_size_cstrs.precedence iof_and_size_sol &&
  check_linear_cstrs iof_and_size_cstrs.periodicity iof_and_size_sol &&
  check_linear_cstrs iof_and_size_cstrs.sufficient_indexes iof_and_size_sol &&
  check_linear_cstrs iof_and_size_cstrs.increase_indexes iof_and_size_sol

(* ======================================================================= *)
(** {2 Word reconstruction} *)


(** Compute the prefix size of a variable c_n. It is the index before
    the first 1 in c_n.v.
*)
let compute_size_c_n_u c_n nbones_c_n_u iof_and_size_sol =
  let iof_c_n j = Integer_var_map.find (Iof (c_n, j)) iof_and_size_sol in
  let rec find j =
    try iof_c_n j
    with Not_found -> iof_c_n (j + 1) - 1
  in
  find (nbones_c_n_u + 1) - 1


(** Build an ultimately periodic binary word from the number of 1s of
    its prefix and periodic part, the size of its prefix and periodic part,
    and the indexes of its 1s.
*)
let build_pbword c_n nbones_c_n_u nbones_c_n_v iof_and_size_sol =
  let size_c_n_u = compute_size_c_n_u c_n nbones_c_n_u iof_and_size_sol in
  let size_c_n_v =
    try Integer_var_map.find (Size c_n) iof_and_size_sol
    with Not_found -> assert false
  in
  let iof_c_n j = Integer_var_map.find (Iof (c_n, j)) iof_and_size_sol in
  let rec push_zero n u =
    assert (n >= 0);
    if n = 0 then u
    else push_zero (n - 1) (false :: u)
  in
  let rec push_one n u =
    assert (n >= 0);
    if n = 0 then u
    else push_one (n - 1) (true :: u)
  in
  let rec build curr_i j u v =
    if j = 0 then
      { u = push_zero curr_i u;
        v = v; }
    else
      let i =
        try iof_c_n j
        with Not_found -> curr_i
      in
      if j > nbones_c_n_u then
        build (i - 1) (j - 1)
          u (push_one 1 (push_zero (curr_i - i) v))
      else
        build (i - 1) (j - 1)
          (push_one 1 (push_zero (curr_i - i) u)) v
  in
  build
    (size_c_n_u + size_c_n_v) (nbones_c_n_u + nbones_c_n_v)
    [] []


(** Build the ultimately periodic binary words corresponding to the
    solution of sizes and indexes of 1s.
*)
let pbwords_of_iof_and_size_sol choice iof_and_size_sol =
  Word_var_map.mapi
    (fun c_n (nbones_c_n_u, nbones_c_n_v) ->
      build_pbword c_n nbones_c_n_u nbones_c_n_v iof_and_size_sol)
    choice.nbones_env

(** Check the solution of an adaptability constraint system. *)
let check_adaptability_sol adaptability_sol adaptability_cstrs =
  let rec subst e sol =
    match e with
    | Var c ->
        begin try Pbw (Word_var_map.find c sol)
        with Not_found -> assert false
        end
    | Pbw p -> Pbw p
    | On (e1, e2) -> On (subst e1 sol, subst e2 sol)
  in
  let adaptability_cstrs =
    List.map
      (fun (Adaptability (e1, e2)) ->
        Adaptability (subst e1 adaptability_sol, subst e2 adaptability_sol))
      adaptability_cstrs
  in
  let simplified_adaptability_cstrs =
    simplified_adaptability_cstrs_of_adaptability_cstrs adaptability_cstrs
  in
  simplified_adaptability_cstrs = []


(* ======================================================================= *)
(** {2 Subtyping constraints solving} *)

(** Solve a subtyping constraint system. *)
let constraint_system_solving ?(verbose = false) k k' subtyping_cstrs =
  if verbose then begin
    Format.printf "Subtyping constraints:@\n  @[%a@]@\n"
      fprint_subtyping_cstrs subtyping_cstrs
  end;
  let adaptability_cstrs =
    adaptability_cstrs_of_subtyping_cstrs subtyping_cstrs
  in
  if verbose then begin
    Format.printf "Adaptability constraints:@\n  @[%a@]@\n"
      fprint_adaptability_cstrs adaptability_cstrs
  end;
  let iof_and_size_cstrs =
    iof_and_size_cstrs_of_adaptability_cstrs ~verbose:verbose
      k k' adaptability_cstrs
  in
  if verbose then begin
    Format.printf "Indexes of 1s and size constraints:@\n  @[";
    Format.printf "Simplified synchronizability constraints:@\n  @[%a@]@\n"
      fprint_linear_cstrs iof_and_size_cstrs.synchronizability;
    Format.printf "Simplified precedence constraints:@\n  @[%a@]@\n"
      fprint_linear_cstrs iof_and_size_cstrs.precedence;
    Format.printf "Periodicity constraints:@\n  @[%a@]@\n"
      fprint_linear_cstrs iof_and_size_cstrs.periodicity;
    Format.printf "Sufficent size constraints:@\n  @[%a@]@\n"
      fprint_linear_cstrs iof_and_size_cstrs.sufficient_size;
    Format.printf "Sufficent indexes constraints:@\n  @[%a@]@\n"
      fprint_linear_cstrs iof_and_size_cstrs.sufficient_indexes;
    Format.printf "Increasing indexes:@\n  @[%a@]@\n"
      fprint_linear_cstrs iof_and_size_cstrs.increase_indexes;
    Format.printf "@]@\n"
  end;
  let iof_and_size_sol =
    iof_and_size_cstrs_solving ~verbose:verbose iof_and_size_cstrs
  in
  assert (check_iof_and_size_sol iof_and_size_cstrs iof_and_size_sol);
  let adaptability_sol =
    pbwords_of_iof_and_size_sol iof_and_size_cstrs.choice iof_and_size_sol
  in
  if verbose then begin
    Format.printf "Solution of adaptability constraints:@\n  @[";
    Word_var_map.iter
      (fun c_n p ->
        Format.printf "%a = %a@\n" fprint_word_var c_n fprint_pbw p)
      adaptability_sol;
    Format.printf "@]@\n"
  end;
  assert (check_adaptability_sol adaptability_sol adaptability_cstrs);
  adaptability_sol

(* ======================================================================= *)
(** {2 Example}

    Let us transform the following subtyping constraint system
    into linear constraints:
    [{ 'a_1 on 10(1)  <: 'a_2 on (01) ;
       'a_2 on (01) on (10) <: 'a_3 on (1) ;
       'a_2 on (01) on (01) <: 'a_3 on (1) ; }].

*)

let subtyping_cstrs_C' =
  let p_10_1 = { u = [ true; false; ]; v = [ true; ]; } in
  let p__1100 = { u = []; v = [ true; true; false; false; ]; } in
  let p__10 = { u = []; v = [ true; false; ]; } in
  let p__01 = { u = []; v = [ false; true; ]; } in
  [ Subtype (CT_on (CT_var (Alpha 1), p_10_1),
             CT_on (CT_var (Alpha 2), p__01));
    Subtype (CT_on (CT_var (Alpha 2), p__1100),
             CT_on (CT_var (Alpha 2), p__01));
    Subtype (CT_on (CT_on (CT_var (Alpha 2), p__01), p__10),
             CT_var (Alpha 3));
    Subtype (CT_on (CT_on (CT_var (Alpha 2), p__01), p__01),
             CT_var (Alpha 3)); ]

let iof_and_size_cstrs_C' =
  constraint_system_solving ~verbose:true 0 1 subtyping_cstrs_C'
